/**
 * Native bindings for the Beeper Client SDK, providing a JavaScript interface
 * to the underlying native implementation.
 */

const path = require('path')
const { binary } = require('../package.json')

/**
 * Configuration options for initializing a Beeper client
 * @typedef {Object} BeeperClientOptions
 * @property {string} dataDirPath - Path to the data directory for storing client data
 * @property {string} deviceName - Name of the device running the client
 * @property {string} appVersion - Version of the application
 * @property {string} deviceModel - Model identifier of the device
 * @property {string} deviceType - Type of device running the client
 * @property {Function} callback - Callback function for client events
 */

const { BeeperClient: BeeperClientNative } = require(path.join(__dirname, 'lib',
	`napi-v${binary.napi_versions}`, 'beeper_client_sdk_napi_binding.node'))

const SUBOBJECT_MAP = {
	'__GetMatrix_': 'matrix',
	'__GetBridges_': 'bridges',
}

/**
 * Custom error class for wrapping native SDK errors with improved stack traces
 */
class SDKError extends Error {
	/**
	 * @param {Error|Object} nativeError - The original error from the native binding or an SDKError object
	 * @param {string} methodName - The name of the method where the error occurred
	 * @throws {TypeError} If nativeError is not an instance of Error or doesn't have required properties
	 */
	constructor(nativeError, methodName) {
		// Handle both standard errors and SDKError objects from the native side
		if (nativeError && typeof nativeError === 'object' && nativeError.internalError !== undefined) {
			// This is an SDKError object from the native side
			super(nativeError.displayError || nativeError.internalError)
			this.name = 'SDKError'
			this.internalError = nativeError.internalError
			this.displayError = nativeError.displayError
			this.canRetry = !!nativeError.canRetry

			// Ensure there's a user-friendly message
			this.message = this.displayError || this.internalError
			if (methodName) {
				this.message = `${methodName} failed: ${this.message}`
			}
		} else if (nativeError instanceof Error) {
			// Standard error from the native binding
			super(nativeError.message)
			this.name = 'SDKError'
			this.message = methodName ? `${methodName} failed: ${nativeError.message}` : nativeError.message
			this.cause = nativeError
			this.internalError = nativeError.message
			this.displayError = ''
			this.canRetry = false
		} else {
			throw new TypeError('SDKError is only constructible from an upstream Error or SDKError object')
		}

		// maintain stack trace information (V8 only)
		Error.captureStackTrace?.(this, SDKError)
	}
}

/**
 * Main client class for interacting with the Beeper SDK. Provides a JavaScript interface
 * to the native implementation with automatic error handling and promise support.
 */
class BeeperClient {
	/**
	 * @param {BeeperClientOptions} options - Configuration options for the client
	 */
	constructor(options) {
		this.__native = new BeeperClientNative(
			options.dataDirPath,
			options.deviceName,
			options.appVersion,
			options.deviceModel,
			options.deviceType,
			options.callback,
		)

		for (let sub of Object.values(SUBOBJECT_MAP)) {
			this[sub] = {}
		}

		for (let methodName in this.__native) {
			if (methodName in BeeperClient.prototype) {
				continue;
			}

			if (methodName.startsWith('__')) {
				for (let sub in SUBOBJECT_MAP) {
					if (methodName.startsWith(sub)) {
						const newMethodName = methodName.replace(sub, '')
						this[SUBOBJECT_MAP[sub]][newMethodName] = this.wrapMethod(methodName)
					}
				}
			} else {
				BeeperClient.prototype[methodName] = this.wrapMethod(methodName)
			}
		}
	}

	/**
	 * Wraps native method calls with error handling and promise support.
	 * Automatically catches and wraps both synchronous errors and promise rejections.
	 * @private
	 */
	wrapMethod = (methodName) => (...args) => {
	// errors created in go will lack stack trace information, which isn't
	// very fun. wrap thrown errors and promise rejections with our own
	// subclass, which also permits effective runtime type tests via
	// instanceof

		try {
			const returnValue = this.__native[methodName](...args)

			if (returnValue?.then) {
				return returnValue.catch(error => {
					throw new SDKError(error, methodName)
				})
			}

			return returnValue
		} catch (error) {
			throw new SDKError(error, methodName)
		}
	}
}

module.exports = { BeeperClient, SDKError }
